import sys
from typing import Any, Coroutine, TypeVar

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

import asyncio
import threading

from asyncer import asyncify
from syncer import sync
from chainlit.context import context

make_async = asyncify

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")
T = TypeVar("T")


def run_sync(co: Coroutine[Any, Any, T_Retval]) -> T_Retval:
    """Run the coroutine synchronously."""
    # Execute from the main thread in the main event loop
    if threading.current_thread() == threading.main_thread():
        return sync(co)
    else:  # Execute from a thread in the main event loop
        result = asyncio.run_coroutine_threadsafe(co, loop=context.loop)
        return result.result()



import json
import uuid
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Dict, List, Optional, Union, cast

from chainlit.action import Action
from chainlit.client.base import MessageDict
from chainlit.config import config
from chainlit.context import context
from chainlit.data import chainlit_client
from chainlit.element import ElementBased
from chainlit.logger import logger
from chainlit.prompt import Prompt
from chainlit.telemetry import trace_event
from chainlit.types import (
    AskActionResponse,
    AskActionSpec,
    AskFileResponse,
    AskFileSpec,
    AskResponse,
    AskSpec,
)
from syncer import asyncio


class MessageBase(ABC):
    id: str
    author: str
    content: str = ""
    streaming = False
    created_at: Union[int, str, None] = None
    fail_on_persist_error: bool = False
    persisted = False

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        if not getattr(self, "id", None):
            self.id = str(uuid.uuid4())
        if not self.created_at:
            self.created_at = datetime.now(timezone.utc).isoformat()

    @abstractmethod
    def to_dict(self) -> Dict:
        pass

    async def with_conversation_id(self):
        _dict = self.to_dict()
        _dict["conversationId"] = await context.session.get_conversation_id()
        return _dict

    async def _create(self):
        msg_dict = await self.with_conversation_id()
        asyncio.create_task(self._persist_create(msg_dict))

        if not config.features.prompt_playground:
            msg_dict.pop("prompt", None)
        return msg_dict

    async def _persist_create(self, message: MessageDict):
        if not chainlit_client or self.persisted:
            return

        try:
            persisted_id = await chainlit_client.create_message(message)

            if persisted_id:
                self.id = persisted_id
                self.persisted = True
        except Exception as e:
            if self.fail_on_persist_error:
                raise e
            logger.error(f"Failed to persist message creation: {str(e)}")

    async def update(
        self,
    ):
        """
        Update a message already sent to the UI.
        """
        trace_event("update_message")
        if self.streaming:
            self.streaming = False
        msg_dict = self.to_dict()
        asyncio.create_task(self._persist_update(msg_dict))
        await context.emitter.update_message(msg_dict)

        return True

    async def _persist_update(self, message: MessageDict):
        if not chainlit_client or not self.id:
            return

        try:
            await chainlit_client.update_message(self.id, message)
        except Exception as e:
            if self.fail_on_persist_error:
                raise e
            logger.error(f"Failed to persist message update: {str(e)}")

    async def remove(self):
        """
        Remove a message already sent to the UI.
        This will not automatically remove potential nested messages and could lead to undesirable side effects in the UI.
        """
        trace_event("remove_message")

        if chainlit_client and self.id:
            await chainlit_client.delete_message(self.id)

        await context.emitter.delete_message(self.to_dict())

        return True

    async def _persist_remove(self):
        if not chainlit_client or not self.id:
            return

        try:
            await chainlit_client.delete_message(self.id)
        except Exception as e:
            if self.fail_on_persist_error:
                raise e
            logger.error(f"Failed to persist message deletion: {str(e)}")

    async def send(self):
        if self.content is None:
            self.content = ""

        if config.code.author_rename:
            self.author = await config.code.author_rename(self.author)

        if self.streaming:
            self.streaming = False

        msg_dict = await self._create()

        await context.emitter.send_message(msg_dict)

        return self.id

    async def stream_token(self, token: str, is_sequence=False):
        """
        Sends a token to the UI. This is useful for streaming messages.
        Once all tokens have been streamed, call .send() to end the stream and persist the message if persistence is enabled.
        """

        if not self.streaming:
            self.streaming = True
            msg_dict = self.to_dict()
            await context.emitter.stream_start(msg_dict)

        if is_sequence:
            self.content = token
        else:
            self.content += token

        assert self.id
        await context.emitter.send_token(
            id=self.id, token=token, is_sequence=is_sequence
        )


class Message(MessageBase):
    """
    Send a message to the UI
    If a project ID is configured, the message will be persisted in the cloud.

    Args:
        content (Union[str, Dict]): The content of the message.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        prompt (Prompt, optional): The prompt used to generate the message. If provided, enables the prompt playground for this message.
        language (str, optional): Language of the code is the content is code. See https://react-code-blocks-rajinwonderland.vercel.app/?path=/story/codeblock--supported-languages for a list of supported languages.
        parent_id (str, optional): If provided, the message will be nested inside the parent in the UI.
        indent (int, optional): If positive, the message will be nested in the UI. (deprecated, use parent_id instead)
        actions (List[Action], optional): A list of actions to send with the message.
        elements (List[ElementBased], optional): A list of elements to send with the message.
        disable_human_feedback (bool, optional): Hide the feedback buttons for this specific message
    """

    def __init__(
        self,
        content: Union[str, Dict],
        author: str = config.ui.name,
        prompt: Optional[Prompt] = None,
        language: Optional[str] = None,
        parent_id: Optional[str] = None,
        indent: int = 0,
        actions: Optional[List[Action]] = None,
        elements: Optional[List[ElementBased]] = None,
        disable_human_feedback: Optional[bool] = False,
        author_is_user: Optional[bool] = False,
        id: Optional[uuid.UUID] = None,
    ):
        self.language = language

        if isinstance(content, dict):
            try:
                self.content = json.dumps(content, indent=4, ensure_ascii=False)
                self.language = "json"
            except TypeError:
                self.content = str(content)
                self.language = "python"
        elif isinstance(content, str):
            self.content = content
        else:
            self.content = str(content)
            self.language = "python"

        if id:
            self.id = str(id)

        self.author = author
        self.author_is_user = author_is_user
        self.prompt = prompt
        self.parent_id = parent_id
        self.indent = indent
        self.actions = actions if actions is not None else []
        self.elements = elements if elements is not None else []
        self.disable_human_feedback = disable_human_feedback

        super().__post_init__()

    @classmethod
    def from_dict(self, _dict: MessageDict):
        message = Message(
            content=_dict["content"],
            author=_dict.get("author", config.ui.name),
            prompt=_dict.get("prompt"),
            language=_dict.get("language"),
            parent_id=_dict.get("parentId"),
            indent=_dict.get("indent") or 0,
            disable_human_feedback=_dict.get("disableHumanFeedback"),
            author_is_user=_dict.get("authorIsUser"),
        )

        if _id := _dict.get("id"):
            message.id = _id
        if created_at := _dict.get("createdAt"):
            message.created_at = created_at

        return message

    def to_dict(self):
        _dict = {
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "authorIsUser": self.author_is_user,
            "language": self.language,
            "parentId": self.parent_id,
            "indent": self.indent,
            "streaming": self.streaming,
            "disableHumanFeedback": self.disable_human_feedback,
        }

        if self.prompt:
            _dict["prompt"] = self.prompt.to_dict()

        if self.id:
            _dict["id"] = self.id

        return _dict

    async def send(self) -> str:
        """
        Send the message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_message")
        id = await super().send()

        if not self.parent_id:
            context.session.root_message = self

        # Create tasks for all actions and elements
        tasks = [action.send(for_id=id) for action in self.actions]
        tasks.extend(element.send(for_id=id) for element in self.elements)

        # Run all tasks concurrently
        await asyncio.gather(*tasks)

        return id

    async def update(self):
        """
        Send the message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_message")
        await super().update()

        actions_to_update = [action for action in self.actions if action.forId is None]

        elements_to_update = [el for el in self.elements if self.id not in el.for_ids]

        for action in actions_to_update:
            await action.send(for_id=self.id)

        for element in elements_to_update:
            await element.send(for_id=self.id)

        return True

    async def remove_actions(self):
        for action in self.actions:
            await action.remove()


class ErrorMessage(MessageBase):
    """
    Send an error message to the UI
    If a project ID is configured, the message will be persisted in the cloud.

    Args:
        content (str): Text displayed above the upload button.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        parent_id (str, optional): If provided, the message will be nested inside the parent in the UI.
        indent (int, optional): If positive, the message will be nested in the UI.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        parent_id: Optional[str] = None,
        indent: int = 0,
        fail_on_persist_error: bool = False,
    ):
        self.content = content
        self.author = author
        self.parent_id = parent_id
        self.indent = indent
        self.fail_on_persist_error = fail_on_persist_error

        super().__post_init__()

    def to_dict(self):
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "parentId": self.parent_id,
            "indent": self.indent,
            "isError": True,
        }

    async def send(self):
        """
        Send the error message to the UI and persist it in the cloud if a project ID is configured.
        Return the ID of the message.
        """
        trace_event("send_error_message")
        return await super().send()


class AskMessageBase(MessageBase):
    async def remove(self):
        removed = await super().remove()
        if removed:
            await context.emitter.clear_ask()


class AskUserMessage(AskMessageBase):
    """
    Ask for the user input before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the message will be uploaded to the cloud storage.

    Args:
        content (str): The content of the prompt.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        disable_human_feedback (bool, optional): Hide the feedback buttons for this specific message
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    """

    def __init__(
        self,
        content: str,
        author: str = config.ui.name,
        disable_human_feedback: bool = False,
        timeout: int = 60,
        raise_on_timeout: bool = False,
    ):
        self.content = content
        self.author = author
        self.timeout = timeout
        self.disable_human_feedback = disable_human_feedback
        self.raise_on_timeout = raise_on_timeout

        super().__post_init__()

    def to_dict(self):
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
            "disableHumanFeedback": self.disable_human_feedback,
        }

    async def send(self) -> Union[AskResponse, None]:
        """
        Sends the question to ask to the UI and waits for the reply.
        """
        trace_event("send_ask_user")

        if self.streaming:
            self.streaming = False

        if config.code.author_rename:
            self.author = await config.code.author_rename(self.author)

        msg_dict = await self._create()

        spec = AskSpec(type="text", timeout=self.timeout)

        res = await context.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout)

        return res


class AskFileMessage(AskMessageBase):
    """
    Ask the user to upload a file before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    If a project ID is configured, the file will be uploaded to the cloud storage.

    Args:
        content (str): Text displayed above the upload button.
        accept (Union[List[str], Dict[str, List[str]]]): List of mime type to accept like ["text/csv", "application/pdf"] or a dict like {"text/plain": [".txt", ".py"]}.
        max_size_mb (int, optional): Maximum size per file in MB. Maximum value is 100.
        max_files (int, optional): Maximum number of files to upload. Maximum value is 10.
        author (str, optional): The author of the message, this will be used in the UI. Defaults to the chatbot name (see config).
        disable_human_feedback (bool, optional): Hide the feedback buttons for this specific message
        timeout (int, optional): The number of seconds to wait for an answer before raising a TimeoutError.
        raise_on_timeout (bool, optional): Whether to raise a socketio TimeoutError if the user does not answer in time.
    """

    def __init__(
        self,
        content: str,
        accept: Union[List[str], Dict[str, List[str]]],
        max_size_mb=2,
        max_files=1,
        author=config.ui.name,
        disable_human_feedback: bool = False,
        timeout=90,
        raise_on_timeout=False,
    ):
        self.content = content
        self.max_size_mb = max_size_mb
        self.max_files = max_files
        self.accept = accept
        self.author = author
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timeout
        self.disable_human_feedback = disable_human_feedback

        super().__post_init__()

    def to_dict(self):
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
            "disableHumanFeedback": self.disable_human_feedback,
        }

    async def send(self) -> Union[List[AskFileResponse], None]:
        """
        Sends the message to request a file from the user to the UI and waits for the reply.
        """
        trace_event("send_ask_file")

        if self.streaming:
            self.streaming = False

        if config.code.author_rename:
            self.author = await config.code.author_rename(self.author)

        msg_dict = await self._create()

        spec = AskFileSpec(
            type="file",
            accept=self.accept,
            max_size_mb=self.max_size_mb,
            max_files=self.max_files,
            timeout=self.timeout,
        )

        res = await context.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout)

        if res:
            return [AskFileResponse(**r) for r in res]
        else:
            return None


class AskActionMessage(AskMessageBase):
    """
    Ask the user to select an action before continuing.
    If the user does not answer in time (see timeout), a TimeoutError will be raised or None will be returned depending on raise_on_timeout.
    """

    def __init__(
        self,
        content: str,
        actions: List[Action],
        author=config.ui.name,
        disable_human_feedback=False,
        timeout=90,
        raise_on_timeout=False,
    ):
        self.content = content
        self.actions = actions
        self.author = author
        self.disable_human_feedback = disable_human_feedback
        self.timeout = timeout
        self.raise_on_timeout = raise_on_timeout

        super().__post_init__()

    def to_dict(self):
        return {
            "id": self.id,
            "createdAt": self.created_at,
            "content": self.content,
            "author": self.author,
            "waitForAnswer": True,
            "disableHumanFeedback": self.disable_human_feedback,
            "timeout": self.timeout,
            "raiseOnTimeout": self.raise_on_timeout,
        }

    async def send(self) -> Union[AskActionResponse, None]:
        """
        Sends the question to ask to the UI and waits for the reply
        """
        trace_event("send_ask_action")

        if self.streaming:
            self.streaming = False

        if config.code.author_rename:
            self.author = await config.code.author_rename(self.author)

        msg_dict = await self._create()
        action_keys = []

        for action in self.actions:
            action_keys.append(action.id)
            await action.send(for_id=str(msg_dict["id"]))

        spec = AskActionSpec(type="action", timeout=self.timeout, keys=action_keys)

        res = cast(
            Union[AskActionResponse, None],
            await context.emitter.send_ask_user(msg_dict, spec, self.raise_on_timeout),
        )

        for action in self.actions:
            await action.remove()
        if res is None:
            self.content = "Timed out: no action was taken"
        else:
            self.content = f'**Selected action:** {res["label"]}'
        await self.update()

        return res



from chainlit.data import chainlit_client
from fastapi import HTTPException


async def is_conversation_author(username: str, conversation_id: str):
    if not chainlit_client:
        raise HTTPException(status_code=401, detail="Unauthorized")

    conversation_author = await chainlit_client.get_conversation_author(conversation_id)

    if conversation_author != username:
        raise HTTPException(status_code=401, detail="Unauthorized")
    else:
        return True



import os
from typing import Optional

from chainlit.client.cloud import ChainlitCloudClient
from chainlit.config import config

chainlit_client = None  # type: Optional[ChainlitCloudClient]

if config.data_persistence:
    chainlit_client = ChainlitCloudClient(
        api_key=os.environ.get("CHAINLIT_API_KEY", ""),
        chainlit_server=config.chainlit_server,
    )



# This is a simple example of a chainlit app.

from chainlit import AskUserMessage, Message, on_chat_start


@on_chat_start
async def main():
    res = await AskUserMessage(content="What is your name?", timeout=30).send()
    if res:
        await Message(
            content=f"Your name is: {res['content']}.\nChainlit installation is working!\nYou can now start building your own chainlit apps!",
        ).send()



import asyncio
import uuid
from typing import Any, Dict, Optional

from chainlit.client.base import ConversationDict, MessageDict
from chainlit.element import Element
from chainlit.message import Message
from chainlit.session import BaseSession, WebsocketSession
from chainlit.types import AskSpec, UIMessagePayload
from socketio.exceptions import TimeoutError


class BaseChainlitEmitter:
    """
    Chainlit Emitter Stub class. This class is used for testing purposes.
    It stubs the ChainlitEmitter class and does nothing on function calls.
    """

    session: BaseSession

    def __init__(self, session: BaseSession) -> None:
        """Initialize with the user session."""
        self.session = session

    async def emit(self, event: str, data: Any):
        """Stub method to get the 'emit' property from the session."""
        pass

    async def ask_user(self):
        """Stub method to get the 'ask_user' property from the session."""
        pass

    async def resume_conversation(self, conv_dict: ConversationDict):
        """Stub method to resume a conversation."""
        pass

    async def send_message(self, msg_dict: dict):
        """Stub method to send a message to the UI."""
        pass

    async def update_message(self, msg_dict: dict):
        """Stub method to update a message in the UI."""
        pass

    async def delete_message(self, msg_dict: dict):
        """Stub method to delete a message in the UI."""
        pass

    async def send_ask_timeout(self):
        """Stub method to send a prompt timeout message to the UI."""
        pass

    async def clear_ask(self):
        """Stub method to clear the prompt from the UI."""
        pass

    async def init_conversation(self, msg_dict: MessageDict):
        """Signal the UI that a new conversation (with a user message) exists"""
        pass

    async def process_user_message(self, payload: UIMessagePayload) -> Message:
        """Stub method to process user message."""
        return Message(content="")

    async def send_ask_user(self, msg_dict: dict, spec, raise_on_timeout=False):
        """Stub method to send a prompt to the UI and wait for a response."""
        pass

    async def update_token_count(self, count: int):
        """Stub method to update the token count for the UI."""
        pass

    async def task_start(self):
        """Stub method to send a task start signal to the UI."""
        pass

    async def task_end(self):
        """Stub method to send a task end signal to the UI."""
        pass

    async def stream_start(self, msg_dict: dict):
        """Stub method to send a stream start signal to the UI."""
        pass

    async def send_token(self, id: str, token: str, is_sequence=False):
        """Stub method to send a message token to the UI."""
        pass

    async def set_chat_settings(self, settings: dict):
        """Stub method to set chat settings."""
        pass


class ChainlitEmitter(BaseChainlitEmitter):
    """
    Chainlit Emitter class. The Emitter is not directly exposed to the developer.
    Instead, the developer interacts with the Emitter through the methods and classes exposed in the __init__ file.
    """

    session: WebsocketSession

    def __init__(self, session: WebsocketSession) -> None:
        """Initialize with the user session."""
        self.session = session

    def _get_session_property(self, property_name: str, raise_error=True):
        """Helper method to get a property from the session."""
        if not hasattr(self, "session") or not hasattr(self.session, property_name):
            if raise_error:
                raise ValueError(f"Session does not have property '{property_name}'")
            else:
                return None
        return getattr(self.session, property_name)

    @property
    def emit(self):
        """Get the 'emit' property from the session."""

        return self._get_session_property("emit")

    @property
    def ask_user(self):
        """Get the 'ask_user' property from the session."""
        return self._get_session_property("ask_user")

    def resume_conversation(self, conv_dict: ConversationDict):
        """Send a conversation to the UI to resume it"""
        return self.emit("resume_conversation", conv_dict)

    def send_message(self, msg_dict: Dict):
        """Send a message to the UI."""
        return self.emit("new_message", msg_dict)

    def update_message(self, msg_dict: Dict):
        """Update a message in the UI."""

        return self.emit("update_message", msg_dict)

    def delete_message(self, msg_dict):
        """Delete a message in the UI."""

        return self.emit("delete_message", msg_dict)

    def send_ask_timeout(self):
        """Send a prompt timeout message to the UI."""

        return self.emit("ask_timeout", {})

    def clear_ask(self):
        """Clear the prompt from the UI."""

        return self.emit("clear_ask", {})

    def init_conversation(self, message: MessageDict):
        """Signal the UI that a new conversation (with a user message) exists"""

        return self.emit("init_conversation", message)

    async def process_user_message(self, payload: UIMessagePayload):
        message_dict = payload["message"]
        files = payload["files"]
        # Temporary UUID generated by the frontend should use v4
        assert uuid.UUID(message_dict["id"]).version == 4

        message = Message.from_dict(message_dict)

        asyncio.create_task(message._create())

        if files:
            file_elements = [Element.from_dict(file) for file in files]
            message.elements = file_elements

            async def send_elements():
                for element in message.elements:
                    await element.send(for_id=message.id)

            asyncio.create_task(send_elements())

        if not self.session.has_user_message:
            self.session.has_user_message = True
            await self.init_conversation(await message.with_conversation_id())

        self.session.root_message = message

        return message

    async def send_ask_user(
        self, msg_dict: Dict, spec: AskSpec, raise_on_timeout=False
    ):
        """Send a prompt to the UI and wait for a response."""

        try:
            # Send the prompt to the UI
            res = await self.ask_user(
                {"msg": msg_dict, "spec": spec.to_dict()}, spec.timeout
            )  # type: Optional["MessageDict"]

            # End the task temporarily so that the User can answer the prompt
            await self.task_end()

            if res:
                # If cloud is enabled, store the response in the database/S3
                if spec.type == "text":
                    await self.process_user_message({"message": res, "files": None})
                elif spec.type == "file":
                    # TODO: upload file to S3
                    pass

            await self.clear_ask()
            return res
        except TimeoutError as e:
            await self.send_ask_timeout()

            if raise_on_timeout:
                raise e
        finally:
            await self.task_start()

    def update_token_count(self, count: int):
        """Update the token count for the UI."""

        return self.emit("token_usage", count)

    def task_start(self):
        """
        Send a task start signal to the UI.
        """
        return self.emit("task_start", {})

    def task_end(self):
        """Send a task end signal to the UI."""
        return self.emit("task_end", {})

    def stream_start(self, msg_dict: Dict):
        """Send a stream start signal to the UI."""
        return self.emit(
            "stream_start",
            msg_dict,
        )

    def send_token(self, id: str, token: str, is_sequence=False):
        """Send a message token to the UI."""
        return self.emit(
            "stream_token", {"id": id, "token": token, "isSequence": is_sequence}
        )

    def set_chat_settings(self, settings: Dict[str, Any]):
        self.session.chat_settings = settings



from chainlit.utils import check_module_version

if not check_module_version("haystack", "1.18.0"):
    raise ValueError(
        "Expected Haystack version >= 1.18.0. Run `pip install farm-haystack --upgrade`"
    )



import os
from typing import Any, Dict, List, Union

from chainlit.config import config
from chainlit.prompt import Prompt, PromptMessage
from chainlit.telemetry import trace_event
from chainlit.types import CompletionRequest
from fastapi import HTTPException
from pydantic.dataclasses import dataclass

from chainlit import input_widget


@dataclass
class BaseProvider:
    id: str
    name: str
    env_vars: Dict[str, str]
    inputs: List[input_widget.InputWidget]
    is_chat: bool

    # Format the message based on the template provided
    def format_message(self, message: PromptMessage, prompt: Prompt):
        if message.template:
            message.formatted = self._format_template(message.template, prompt)
        return message

    # Convert the message to string format
    def message_to_string(self, message: PromptMessage):
        return message.formatted

    # Concatenate multiple messages with a joiner
    def concatenate_messages(self, messages: List[PromptMessage], joiner="\n\n"):
        return joiner.join([self.message_to_string(m) for m in messages])

    # Format the template based on the prompt inputs
    def _format_template(self, template: str, prompt: Prompt):
        if prompt.template_format == "f-string":
            return template.format(**(prompt.inputs or {}))
        raise HTTPException(
            status_code=422, detail=f"Unsupported format {prompt.template_format}"
        )

    # Create a prompt based on the request
    def create_prompt(self, request: CompletionRequest):
        prompt = request.prompt
        if prompt.messages:
            messages = [self.format_message(m, prompt=prompt) for m in prompt.messages]
        else:
            messages = None

        if self.is_chat:
            if messages:
                return messages
            elif prompt.template or prompt.formatted:
                return [
                    self.format_message(
                        PromptMessage(
                            template=prompt.template,
                            formatted=prompt.formatted,
                            role="user",
                        ),
                        prompt=prompt,
                    )
                ]
            else:
                raise HTTPException(status_code=422, detail="Could not create prompt")
        else:
            if prompt.template:
                return self._format_template(prompt.template, prompt=prompt)
            elif messages:
                return self.concatenate_messages(messages)
            elif prompt.formatted:
                return prompt.formatted
            else:
                raise HTTPException(status_code=422, detail="Could not create prompt")

    # Create a completion event
    async def create_completion(self, request: CompletionRequest):
        trace_event("completion")

    # Get the environment variable based on the request
    def get_var(self, request: CompletionRequest, var: str) -> Union[str, None]:
        user_env = config.project.user_env or []

        if var in user_env:
            return request.userEnv.get(var)
        else:
            return os.environ.get(var)

    # Check if the environment variable is available
    def _is_env_var_available(self, var: str) -> bool:
        user_env = config.project.user_env or []
        return var in os.environ or var in user_env

    # Check if the provider is configured
    def is_configured(self):
        for var in self.env_vars.values():
            if not self._is_env_var_available(var):
                return False
        return True

    # Validate the environment variables in the request
    def validate_env(self, request: CompletionRequest):
        return {k: self.get_var(request, v) for k, v in self.env_vars.items()}

    # Check if the required settings are present
    def require_settings(self, settings: Dict[str, Any]):
        for _input in self.inputs:
            if _input.id not in settings:
                raise HTTPException(
                    status_code=422,
                    detail=f"Field {_input.id} is a required setting but is not found.",
                )

    # Convert the provider to dictionary format
    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "inputs": [input_widget.to_dict() for input_widget in self.inputs],
            "is_chat": self.is_chat,
        }



from chainlit.utils import check_module_version

if not check_module_version("llama_index", "0.8.3"):
    raise ValueError(
        "Expected LlamaIndex version >= 0.8.3. Run `pip install llama_index --upgrade`"
    )



from typing import (
    Any,
    Dict,
    Generic,
    List,
    Literal,
    Mapping,
    Optional,
    TypedDict,
    TypeVar,
)

from chainlit.logger import logger
from chainlit.prompt import Prompt
from dataclasses_json import DataClassJsonMixin
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass
from python_graphql_client import GraphqlClient

ElementType = Literal[
    "image", "avatar", "text", "pdf", "tasklist", "audio", "video", "file", "plotly"
]
ElementDisplay = Literal["inline", "side", "page"]
ElementSize = Literal["small", "medium", "large"]

Role = Literal["USER", "ADMIN", "OWNER", "ANONYMOUS"]
Provider = Literal[
    "credentials", "header", "github", "google", "azure-ad", "okta", "auth0", "descope"
]


class AppUserDict(TypedDict):
    id: str
    username: str


# Used when logging-in a user
@dataclass
class AppUser(DataClassJsonMixin):
    username: str
    role: Role = "USER"
    tags: List[str] = Field(default_factory=list)
    image: Optional[str] = None
    provider: Optional[Provider] = None


@dataclass
class PersistedAppUserFields:
    id: str
    createdAt: int


@dataclass
class PersistedAppUser(AppUser, PersistedAppUserFields):
    pass


class MessageDict(TypedDict):
    conversationId: Optional[str]
    id: str
    createdAt: Optional[int]
    content: str
    author: str
    prompt: Optional[Prompt]
    language: Optional[str]
    parentId: Optional[str]
    indent: Optional[int]
    authorIsUser: Optional[bool]
    waitForAnswer: Optional[bool]
    isError: Optional[bool]
    humanFeedback: Optional[int]
    disableHumanFeedback: Optional[bool]


class ElementDict(TypedDict):
    id: str
    conversationId: Optional[str]
    type: ElementType
    url: str
    objectKey: Optional[str]
    name: str
    display: ElementDisplay
    size: Optional[ElementSize]
    language: Optional[str]
    forIds: Optional[List[str]]
    mime: Optional[str]


class ConversationDict(TypedDict):
    id: Optional[str]
    metadata: Optional[Dict]
    createdAt: Optional[int]
    appUser: Optional[AppUserDict]
    messages: List[MessageDict]
    elements: Optional[List[ElementDict]]


@dataclass
class PageInfo:
    hasNextPage: bool
    endCursor: Optional[str]


T = TypeVar("T")


@dataclass
class PaginatedResponse(DataClassJsonMixin, Generic[T]):
    pageInfo: PageInfo
    data: List[T]


class Pagination(BaseModel):
    first: int
    cursor: Optional[str] = None


class ConversationFilter(BaseModel):
    feedback: Optional[Literal[-1, 0, 1]] = None
    username: Optional[str] = None
    search: Optional[str] = None


class ChainlitGraphQLClient:
    def __init__(self, api_key: str, chainlit_server: str):
        self.headers = {"content-type": "application/json"}
        if api_key:
            self.headers["x-api-key"] = api_key
        else:
            raise ValueError("Cannot instantiate Cloud Client without CHAINLIT_API_KEY")

        graphql_endpoint = f"{chainlit_server}/api/graphql"
        self.graphql_client = GraphqlClient(
            endpoint=graphql_endpoint, headers=self.headers
        )

    async def query(self, query: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
        Execute a GraphQL query.

        :param query: The GraphQL query string.
        :param variables: A dictionary of variables for the query.
        :return: The response data as a dictionary.
        """
        return await self.graphql_client.execute_async(query=query, variables=variables)

    def check_for_errors(self, response: Dict[str, Any], raise_error: bool = False):
        if "errors" in response:
            if raise_error:
                raise Exception(
                    f"{response['errors'][0]['message']}. Path: {str(response['errors'][0]['path'])}"
                )
            logger.error(response["errors"][0])
            return True
        return False

    async def mutation(
        self, mutation: str, variables: Mapping[str, Any] = {}
    ) -> Dict[str, Any]:
        """
        Execute a GraphQL mutation.

        :param mutation: The GraphQL mutation string.
        :param variables: A dictionary of variables for the mutation.
        :return: The response data as a dictionary.
        """
        return await self.graphql_client.execute_async(
            query=mutation, variables=variables
        )



import os
import threading

from chainlit.config import config
from chainlit.logger import logger


def init_lc_cache():
    use_cache = config.project.cache is True and config.run.no_cache is False

    if use_cache:
        try:
            import langchain
        except ImportError:
            return
        from langchain.cache import SQLiteCache

        if config.project.lc_cache_path is not None:
            langchain.llm_cache = SQLiteCache(
                database_path=config.project.lc_cache_path
            )
            if not os.path.exists(config.project.lc_cache_path):
                logger.info(
                    f"LangChain cache created at: {config.project.lc_cache_path}"
                )


_cache = {}
_cache_lock = threading.Lock()


def cache(func):
    def wrapper(*args, **kwargs):
        # Create a cache key based on the function name, arguments, and keyword arguments
        cache_key = (
            (func.__name__,) + args + tuple((k, v) for k, v in sorted(kwargs.items()))
        )

        with _cache_lock:
            # Check if the result is already in the cache
            if cache_key not in _cache:
                # If not, call the function and store the result in the cache
                _cache[cache_key] = func(*args, **kwargs)

        return _cache[cache_key]

    return wrapper



from typing import Any, Generic, List, Optional, TypeVar

from chainlit.config import config
from chainlit.context import context
from haystack.agents import Agent, Tool
from haystack.agents.agent_step import AgentStep

import chainlit as cl

T = TypeVar("T")


class Stack(Generic[T]):
    def __init__(self) -> None:
        self.items: List[T] = []

    def __len__(self):
        return len(self.items)

    def push(self, item: T) -> None:
        self.items.append(item)

    def pop(self) -> T:
        return self.items.pop()

    def peek(self) -> T:
        return self.items[-1]

    def clear(self) -> None:
        self.items.clear()


class HaystackAgentCallbackHandler:
    stack: Stack[cl.Message]
    latest_agent_message: Optional[cl.Message]

    def __init__(self, agent: Agent):
        agent.callback_manager.on_agent_start += self.on_agent_start
        agent.callback_manager.on_agent_step += self.on_agent_step
        agent.callback_manager.on_agent_finish += self.on_agent_finish
        agent.callback_manager.on_new_token += self.on_new_token

        agent.tm.callback_manager.on_tool_start += self.on_tool_start
        agent.tm.callback_manager.on_tool_finish += self.on_tool_finish
        agent.tm.callback_manager.on_tool_error += self.on_tool_error

    def get_root_message(self):
        if not context.session.root_message:
            root_message = cl.Message(author=config.ui.name, content="")
            cl.run_sync(root_message.send())

        return context.session.root_message

    def on_agent_start(self, **kwargs: Any) -> None:
        # Prepare agent step message for streaming
        self.agent_name = kwargs.get("name", "Agent")
        self.stack = Stack[cl.Message]()
        self.stack.push(self.get_root_message())

        agent_message = cl.Message(
            author=self.agent_name, parent_id=self.stack.peek().id, content=""
        )
        self.stack.push(agent_message)

    # This method is called when a step has finished
    def on_agent_step(self, agent_step: AgentStep, **kwargs: Any) -> None:
        # Send previous agent step message
        self.latest_agent_message = self.stack.pop()

        # If token streaming is disabled
        if self.latest_agent_message.content == "":
            self.latest_agent_message.content = agent_step.prompt_node_response

        cl.run_sync(self.latest_agent_message.send())

        if not agent_step.is_last():
            # Prepare message for next agent step
            agent_message = cl.Message(
                author=self.agent_name, parent_id=self.stack.peek().id, content=""
            )
            self.stack.push(agent_message)

    def on_agent_finish(self, agent_step: AgentStep, **kwargs: Any) -> None:
        self.latest_agent_message = None
        self.stack.clear()

    def on_new_token(self, token, **kwargs: Any) -> None:
        # Stream agent step tokens
        cl.run_sync(self.stack.peek().stream_token(token))

    def on_tool_start(self, tool_input: str, tool: Tool, **kwargs: Any) -> None:
        # Tool started, create message
        parent_id = self.latest_agent_message.id if self.latest_agent_message else None
        tool_message = cl.Message(author=tool.name, parent_id=parent_id, content="")
        self.stack.push(tool_message)

    def on_tool_finish(
        self,
        tool_result: str,
        tool_name: Optional[str] = None,
        tool_input: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        # Tool finished, send message with tool_result
        tool_message = self.stack.pop()
        tool_message.content = tool_result
        cl.run_sync(tool_message.send())

    def on_tool_error(self, exception: Exception, tool: Tool, **kwargs: Any) -> None:
        # Tool error, send error message
        cl.run_sync(self.stack.pop().remove())
        cl.run_sync(cl.ErrorMessage(str(exception), author=tool.name).send())



import asyncio
import json
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Union

if TYPE_CHECKING:
    from chainlit.message import Message
    from chainlit.types import AskResponse

from chainlit.client.cloud import AppUser, PersistedAppUser
from chainlit.data import chainlit_client


class JSONEncoderIgnoreNonSerializable(json.JSONEncoder):
    def default(self, obj):
        try:
            return super(JSONEncoderIgnoreNonSerializable, self).default(obj)
        except TypeError:
            return None


def clean_metadata(metadata: Dict):
    return json.loads(
        json.dumps(metadata, cls=JSONEncoderIgnoreNonSerializable, ensure_ascii=False)
    )


class BaseSession:
    """Base object."""

    def __init__(
        self,
        # Id of the session
        id: str,
        # Logged-in user informations
        user: Optional[Union["AppUser", "PersistedAppUser"]],
        # Logged-in user token
        token: Optional[str],
        # User specific environment variables. Empty if no user environment variables are required.
        user_env: Optional[Dict[str, str]],
        # Last message at the root of the chat
        root_message: Optional["Message"] = None,
        # Chat profile selected before the session was created
        chat_profile: Optional[str] = None,
        # Conversation id to resume
        conversation_id: Optional[str] = None,
    ):
        self.user = user
        self.token = token
        self.root_message = root_message
        self.has_user_message = False
        self.user_env = user_env or {}
        self.chat_profile = chat_profile

        self.id = id
        self.conversation_id = conversation_id

        self.chat_settings: Dict[str, Any] = {}

        self.lock = asyncio.Lock()

    async def get_conversation_id(self) -> Optional[str]:
        if not chainlit_client:
            return None

        if isinstance(self, HTTPSession):
            tags = ["api"]
        else:
            tags = ["chat"]

        async with self.lock:
            if not self.conversation_id:
                app_user_id = (
                    self.user.id if isinstance(self.user, PersistedAppUser) else None
                )
                self.conversation_id = await chainlit_client.create_conversation(
                    app_user_id=app_user_id, tags=tags
                )
        return self.conversation_id

    def to_persistable(self) -> Dict:
        from chainlit.user_session import user_sessions

        user_session = user_sessions.get(self.id) or {}  # type: Dict
        user_session["chat_settings"] = self.chat_settings
        user_session["chat_profile"] = self.chat_profile
        metadata = clean_metadata(user_session)
        return metadata


class HTTPSession(BaseSession):
    """Internal HTTP session object. Used to consume Chainlit through API (no websocket)."""

    def __init__(
        self,
        # Id of the session
        id: str,
        # Logged-in user informations
        user: Optional[Union["AppUser", "PersistedAppUser"]],
        # Logged-in user token
        token: Optional[str],
        user_env: Optional[Dict[str, str]],
        # Last message at the root of the chat
        root_message: Optional["Message"] = None,
        # User specific environment variables. Empty if no user environment variables are required.
    ):
        super().__init__(
            id=id, user=user, token=token, user_env=user_env, root_message=root_message
        )


class WebsocketSession(BaseSession):
    """Internal web socket session object.

    A socket id is an ephemeral id that can't be used as a session id
    (as it is for instance regenerated after each reconnection).

    The Session object store an internal mapping between socket id and
    a server generated session id, allowing to persists session
    between socket reconnection but also retrieving a session by
    socket id for convenience.
    """

    def __init__(
        self,
        # Id from the session cookie
        id: str,
        # Associated socket id
        socket_id: str,
        # Function to emit a message to the user
        emit: Callable[[str, Any], None],
        # Function to ask the user a question
        ask_user: Callable[[Any, Optional[int]], Union["AskResponse", None]],
        # User specific environment variables. Empty if no user environment variables are required.
        user_env: Dict[str, str],
        # Logged-in user informations
        user: Optional[Union["AppUser", "PersistedAppUser"]],
        # Logged-in user token
        token: Optional[str],
        # Last message at the root of the chat
        root_message: Optional["Message"] = None,
        # Chat profile selected before the session was created
        chat_profile: Optional[str] = None,
        # Conversation id to resume
        conversation_id: Optional[str] = None,
    ):
        super().__init__(
            id=id,
            user=user,
            token=token,
            user_env=user_env,
            root_message=root_message,
            chat_profile=chat_profile,
            conversation_id=conversation_id,
        )

        self.socket_id = socket_id
        self.ask_user = ask_user
        self.emit = emit

        self.should_stop = False
        self.restored = False

        ws_sessions_id[self.id] = self
        ws_sessions_sid[socket_id] = self

    def restore(self, new_socket_id: str):
        """Associate a new socket id to the session."""
        ws_sessions_sid.pop(self.socket_id, None)
        ws_sessions_sid[new_socket_id] = self
        self.socket_id = new_socket_id
        self.restored = True

    def delete(self):
        """Delete the session."""
        ws_sessions_sid.pop(self.socket_id, None)
        ws_sessions_id.pop(self.id, None)

    @classmethod
    def get(cls, socket_id: str):
        """Get session by socket id."""
        return ws_sessions_sid.get(socket_id)

    @classmethod
    def get_by_id(cls, session_id: str):
        """Get session by session id."""
        return ws_sessions_id.get(session_id)

    @classmethod
    def require(cls, socket_id: str):
        """Throws an exception if the session is not found."""
        if session := cls.get(socket_id):
            return session
        raise ValueError("Session not found")


ws_sessions_sid: Dict[str, WebsocketSession] = {}
ws_sessions_id: Dict[str, WebsocketSession] = {}



from chainlit.input_widget import Select, Slider, Tags
from chainlit.playground.provider import BaseProvider
from chainlit.prompt import PromptMessage
from fastapi import HTTPException
from fastapi.responses import StreamingResponse


class AnthropicProvider(BaseProvider):
    def message_to_string(self, message: PromptMessage) -> str:
        import anthropic

        if message.role == "user":
            message_text = f"{anthropic.HUMAN_PROMPT} {message.formatted}"
        elif message.role == "assistant":
            message_text = f"{anthropic.AI_PROMPT} {message.formatted}"
        elif message.role == "function":
            message_text = f"{anthropic.AI_PROMPT} {message.formatted}"
        elif message.role == "system":
            message_text = (
                f"{anthropic.HUMAN_PROMPT} <admin>{message.formatted}</admin>"
            )
        else:
            raise HTTPException(status_code=400, detail=f"Got unknown type {message}")
        return message_text

    async def create_completion(self, request):
        await super().create_completion(request)
        import anthropic

        env_settings = self.validate_env(request=request)

        llm_settings = request.prompt.settings
        self.require_settings(llm_settings)

        prompt = self.concatenate_messages(self.create_prompt(request), joiner="")

        if not prompt.endswith(anthropic.AI_PROMPT):
            prompt += anthropic.AI_PROMPT

        client = anthropic.AsyncAnthropic(**env_settings)

        llm_settings["stream"] = True

        try:
            stream = await client.completions.create(prompt=prompt, **llm_settings)
        except anthropic.APIConnectionError as e:
            raise HTTPException(
                status_code=503,
                detail=e.__cause__,
            )
        except anthropic.RateLimitError as e:
            raise HTTPException(
                status_code=429,
            )
        except anthropic.APIStatusError as e:
            raise HTTPException(status_code=e.status_code, detail=e.response)

        async def create_event_stream():
            async for data in stream:
                token = data.completion
                yield token

        return StreamingResponse(create_event_stream())


Anthropic = AnthropicProvider(
    id="anthropic-chat",
    name="Anthropic",
    env_vars={"api_key": "ANTHROPIC_API_KEY"},
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["claude-2", "claude-instant-1"],
            initial_value="claude-2",
        ),
        Slider(
            id="max_tokens_to_sample",
            label="Max Tokens To Sample",
            min=1.0,
            max=100000,
            step=1.0,
            initial=1000,
        ),
        Tags(
            id="stop_sequences",
            label="Stop Sequences",
            initial=[],
        ),
        Slider(
            id="temperature",
            label="Temperature",
            min=0.0,
            max=1.0,
            step=0.01,
            initial=1,
        ),
        Slider(
            id="top_p",
            label="Top P",
            min=0.0,
            max=1.0,
            step=0.01,
            initial=0.7,
        ),
        Slider(
            id="top_k",
            label="Top K",
            min=0.0,
            max=2048.0,
            step=1.0,
            initial=0,
        ),
    ],
    is_chat=True,
)



import logging
import sys

logging.basicConfig(
    level=logging.INFO,
    stream=sys.stdout,
    format="%(asctime)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logging.getLogger("socketio").setLevel(logging.ERROR)
logging.getLogger("engineio").setLevel(logging.ERROR)
logging.getLogger("numexpr").setLevel(logging.ERROR)


logger = logging.getLogger("chainlit")



import asyncio
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from chainlit.context import context_var
from chainlit.message import Message
from chainlit.playground.providers.openai import stringify_function_call
from chainlit.prompt import Prompt, PromptMessage
from langchain.callbacks.tracers.base import BaseTracer
from langchain.callbacks.tracers.schemas import Run
from langchain.schema.messages import BaseMessage
from langchain.schema.output import ChatGenerationChunk, GenerationChunk

DEFAULT_ANSWER_PREFIX_TOKENS = ["Final", "Answer", ":"]


class FinalStreamHelper:
    # The stream we can use to stream the final answer from a chain
    final_stream: Union[Message, None]
    # Should we stream the final answer?
    stream_final_answer: bool = False
    # Token sequence that prefixes the answer
    answer_prefix_tokens: List[str]
    # Ignore white spaces and new lines when comparing answer_prefix_tokens to last tokens? (to determine if answer has been reached)
    strip_tokens: bool

    answer_reached: bool

    def __init__(
        self,
        answer_prefix_tokens: Optional[List[str]] = None,
        stream_final_answer: bool = False,
        force_stream_final_answer: bool = False,
        strip_tokens: bool = True,
    ) -> None:
        # Langchain final answer streaming logic
        if answer_prefix_tokens is None:
            self.answer_prefix_tokens = DEFAULT_ANSWER_PREFIX_TOKENS
        else:
            self.answer_prefix_tokens = answer_prefix_tokens
        if strip_tokens:
            self.answer_prefix_tokens_stripped = [
                token.strip() for token in self.answer_prefix_tokens
            ]
        else:
            self.answer_prefix_tokens_stripped = self.answer_prefix_tokens

        self.last_tokens = [""] * len(self.answer_prefix_tokens)
        self.last_tokens_stripped = [""] * len(self.answer_prefix_tokens)
        self.strip_tokens = strip_tokens
        self.answer_reached = force_stream_final_answer

        # Our own final answer streaming logic
        self.stream_final_answer = stream_final_answer
        self.final_stream = None
        self.has_streamed_final_answer = False

    def _check_if_answer_reached(self) -> bool:
        if self.strip_tokens:
            return self._compare_last_tokens(self.last_tokens_stripped)
        else:
            return self._compare_last_tokens(self.last_tokens)

    def _compare_last_tokens(self, last_tokens: List[str]):
        if last_tokens == self.answer_prefix_tokens_stripped:
            # If tokens match perfectly we are done
            return True
        else:
            # Some LLMs will consider all the tokens of the final answer as one token
            # so we check if any last token contains all answer tokens
            return any(
                [
                    all(
                        answer_token in last_token
                        for answer_token in self.answer_prefix_tokens_stripped
                    )
                    for last_token in last_tokens
                ]
            )

    def _append_to_last_tokens(self, token: str) -> None:
        self.last_tokens.append(token)
        self.last_tokens_stripped.append(token.strip())
        if len(self.last_tokens) > len(self.answer_prefix_tokens):
            self.last_tokens.pop(0)
            self.last_tokens_stripped.pop(0)


class PromptHelper:
    prompt_sequence: List[Prompt]

    def __init__(self) -> None:
        self.prompt_sequence = []

    @property
    def current_prompt(self):
        return self.prompt_sequence[-1] if self.prompt_sequence else None

    def _convert_message_role(self, role: str):
        if "human" in role.lower():
            return "user"
        elif "system" in role.lower():
            return "system"
        elif "function" in role.lower():
            return "function"
        else:
            return "assistant"

    def _convert_message_dict(
        self,
        message: Dict,
        template: Optional[str] = None,
        template_format: Optional[str] = None,
    ):
        class_name = message["id"][-1]
        kwargs = message.get("kwargs", {})
        function_call = kwargs.get("additional_kwargs", {}).get("function_call")
        if function_call:
            content = stringify_function_call(function_call)
        else:
            content = kwargs.get("content", "")
        return PromptMessage(
            name=kwargs.get("name"),
            role=self._convert_message_role(class_name),
            template=template,
            template_format=template_format,
            formatted=content,
        )

    def _convert_message(
        self,
        message: Union[Dict, BaseMessage],
        template: Optional[str] = None,
        template_format: Optional[str] = None,
    ):
        if isinstance(message, dict):
            return self._convert_message_dict(
                message,
            )
        function_call = message.additional_kwargs.get("function_call")
        if function_call:
            content = stringify_function_call(function_call)
        else:
            content = message.content
        return PromptMessage(
            name=getattr(message, "name", None),
            role=self._convert_message_role(message.type),
            template=template,
            template_format=template_format,
            formatted=content,
        )

    def _get_messages(self, serialized: Dict):
        # In LCEL prompts messages are not at the same place
        lcel_messages = serialized.get("kwargs", {}).get(
            "messages", []
        )  # type: List[Dict]
        if lcel_messages:
            return lcel_messages
        else:
            # For chains
            prompt_params = (
                serialized.get("kwargs", {}).get("prompt", {}).get("kwargs", {})
            )
            chain_messages = prompt_params.get("messages", [])  # type: List[Dict]

            return chain_messages

    def _build_prompt(self, serialized: Dict, inputs: Dict):
        messages = self._get_messages(serialized)
        if messages:
            # If prompt is chat, the formatted values will be added in on_chat_model_start
            self._build_chat_template_prompt(messages, inputs)
        else:
            # For completion prompt everything is done here
            self._build_completion_prompt(serialized, inputs)

    def _build_completion_prompt(self, serialized: Dict, inputs: Dict):
        if not serialized:
            return
        kwargs = serialized.get("kwargs", {})
        template = kwargs.get("template")
        template_format = kwargs.get("template_format")
        stringified_inputs = {k: str(v) for (k, v) in inputs.items()}

        if not template:
            return

        self.prompt_sequence.append(
            Prompt(
                template=template,
                template_format=template_format,
                inputs=stringified_inputs,
            )
        )

    def _build_default_prompt(
        self,
        run: Run,
        generation_type: str,
        provider: str,
        llm_settings: Dict,
        completion: str,
    ):
        """Build a prompt once an LLM has been executed if no current prompt exists (without template)"""
        if "chat" in generation_type.lower():
            return Prompt(
                provider=provider,
                settings=llm_settings,
                completion=completion,
                messages=[
                    PromptMessage(
                        formatted=formatted_prompt,
                        role=self._convert_message_role(formatted_prompt.split(":")[0]),
                    )
                    for formatted_prompt in run.inputs.get("prompts", [])
                ],
            )
        else:
            return Prompt(
                provider=provider,
                settings=llm_settings,
                completion=completion,
                formatted=run.inputs.get("prompts", [])[0],
            )

    def _build_chat_template_prompt(self, lc_messages: List[Dict], inputs: Dict):
        def build_template_messages() -> List[PromptMessage]:
            template_messages = []  # type: List[PromptMessage]

            if not lc_messages:
                return template_messages

            for lc_message in lc_messages:
                message_kwargs = lc_message.get("kwargs", {})
                class_name = lc_message["id"][-1]  # type: str
                prompt = message_kwargs.get("prompt", {})
                prompt_kwargs = prompt.get("kwargs", {})
                template = prompt_kwargs.get("template")
                template_format = prompt_kwargs.get("template_format")

                if "placeholder" in class_name.lower():
                    variable_name = lc_message.get(
                        "variable_name"
                    )  # type: Optional[str]
                    variable = inputs.get(variable_name, [])
                    placeholder_size = len(variable)
                    if placeholder_size:
                        template_messages += [
                            PromptMessage(placeholder_size=placeholder_size)
                        ]
                else:
                    template_messages += [
                        PromptMessage(
                            template=template,
                            template_format=template_format,
                            role=self._convert_message_role(class_name),
                        )
                    ]
            return template_messages

        template_messages = build_template_messages()

        if not template_messages:
            return

        stringified_inputs = {k: str(v) for (k, v) in inputs.items()}
        self.prompt_sequence.append(
            Prompt(messages=template_messages, inputs=stringified_inputs)
        )

    def _build_chat_formatted_prompt(
        self, lc_messages: Union[List[BaseMessage], List[dict]]
    ):
        if not self.current_prompt:
            return

        formatted_messages = []  # type: List[PromptMessage]
        if self.current_prompt.messages:
            # This is needed to compute the correct message index to read
            placeholder_offset = 0
            # The final list of messages
            formatted_messages = []
            # Looping the messages built in build_prompt
            # They only contain the template
            for template_index, template_message in enumerate(
                self.current_prompt.messages
            ):
                # If a message has a placeholder size, we need to replace it
                # With the N following messages, where N is the placeholder size
                if template_message.placeholder_size:
                    for _ in range(template_message.placeholder_size):
                        lc_message = lc_messages[template_index + placeholder_offset]
                        formatted_messages += [self._convert_message(lc_message)]
                        # Increment the placeholder offset
                        placeholder_offset += 1
                    # Finally, decrement the placeholder offset by one
                    # Because the message representing the placeholder is now consumed
                    placeholder_offset -= 1
                # The current message is not a placeholder
                else:
                    lc_message = lc_messages[template_index + placeholder_offset]
                    # Update the role and formatted value, keep the template
                    formatted_messages += [
                        self._convert_message(
                            lc_message,
                            template=template_message.template,
                            template_format=template_message.template_format,
                        )
                    ]
            # If the chat llm has more message than the initial chain prompt, append them
            # Typically happens with function agents
            if len(lc_messages) > len(formatted_messages):
                formatted_messages += [
                    self._convert_message(m)
                    for m in lc_messages[len(formatted_messages) :]
                ]
        else:
            formatted_messages = [
                self._convert_message(lc_message) for lc_message in lc_messages
            ]

        self.current_prompt.messages = formatted_messages

    def _build_llm_settings(
        self,
        serialized: Dict,
        invocation_params: Optional[Dict] = None,
    ):
        # invocation_params = run.extra.get("invocation_params")
        if invocation_params is None:
            return None, None

        provider = invocation_params.pop("_type", "")  # type: str

        model_kwargs = invocation_params.pop("model_kwargs", {})

        if model_kwargs is None:
            model_kwargs = {}

        merged = {
            **invocation_params,
            **model_kwargs,
            **serialized.get("kwargs", {}),
        }

        # make sure there is no api key specification
        settings = {k: v for k, v in merged.items() if not k.endswith("_api_key")}

        return provider, settings


DEFAULT_TO_IGNORE = ["RunnableSequence", "RunnableParallel", "<lambda>"]
DEFAULT_TO_KEEP = ["retriever", "llm", "agent", "chain", "tool"]


class LangchainTracer(BaseTracer, PromptHelper, FinalStreamHelper):
    llm_stream_message: Dict[str, Message]
    parent_id_map: Dict[str, str]
    ignored_runs: set

    def __init__(
        self,
        # Token sequence that prefixes the answer
        answer_prefix_tokens: Optional[List[str]] = None,
        # Should we stream the final answer?
        stream_final_answer: bool = False,
        # Should force stream the first response?
        force_stream_final_answer: bool = False,
        # Runs to ignore to enhance readability
        to_ignore: Optional[List[str]] = None,
        # Runs to keep within ignored runs
        to_keep: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        BaseTracer.__init__(self, **kwargs)
        PromptHelper.__init__(self)
        FinalStreamHelper.__init__(
            self,
            answer_prefix_tokens=answer_prefix_tokens,
            stream_final_answer=stream_final_answer,
            force_stream_final_answer=force_stream_final_answer,
        )
        self.context = context_var.get()
        self.llm_stream_message = {}
        self.parent_id_map = {}
        self.ignored_runs = set()
        self.root_parent_id = (
            self.context.session.root_message.id
            if self.context.session.root_message
            else None
        )

        if to_ignore is None:
            self.to_ignore = DEFAULT_TO_IGNORE
        else:
            self.to_ignore = to_ignore

        if to_keep is None:
            self.to_keep = DEFAULT_TO_KEEP
        else:
            self.to_keep = to_keep

    def _run_sync(self, co):
        asyncio.run_coroutine_threadsafe(co, loop=self.context.loop)

    def _persist_run(self, run: Run) -> None:
        pass

    def _get_run_parent_id(self, run: Run):
        parent_id = str(run.parent_run_id) if run.parent_run_id else self.root_parent_id

        return parent_id

    def _get_non_ignored_parent_id(self, current_parent_id: Optional[str] = None):
        if not current_parent_id:
            return self.root_parent_id

        if current_parent_id not in self.parent_id_map:
            return current_parent_id

        while current_parent_id in self.parent_id_map:
            current_parent_id = self.parent_id_map[current_parent_id]

        return current_parent_id

    def _should_ignore_run(self, run: Run):
        parent_id = self._get_run_parent_id(run)

        ignore_by_name = run.name in self.to_ignore
        ignore_by_parent = parent_id in self.ignored_runs

        ignore = ignore_by_name or ignore_by_parent

        if ignore:
            if parent_id:
                # Add the parent id of the ignored run in the mapping
                # so we can re-attach a kept child to the right parent id
                self.parent_id_map[str(run.id)] = parent_id
            # Tag the run as ignored
            self.ignored_runs.add(str(run.id))

        # If the ignore cause is the parent being ignored, check if we should nonetheless keep the child
        if ignore_by_parent and not ignore_by_name and run.run_type in self.to_keep:
            return False, self._get_non_ignored_parent_id(str(run.id))
        else:
            return ignore, parent_id

    def _is_annotable(self, run: Run):
        return run.run_type in ["retriever", "llm"]

    def _get_completion(self, generation: Dict):
        if message := generation.get("message"):
            kwargs = message.get("kwargs", {})
            if function_call := kwargs.get("additional_kwargs", {}).get(
                "function_call"
            ):
                return stringify_function_call(function_call), "json"
            else:
                return kwargs.get("content", ""), None
        else:
            return generation.get("text", ""), None

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Adding formatted content and new message to the previously built template prompt"""
        lc_messages = messages[0]
        if not self.current_prompt:
            self.prompt_sequence.append(
                Prompt(messages=[self._convert_message(m) for m in lc_messages])
            )
        else:
            self._build_chat_formatted_prompt(lc_messages)

        super().on_chat_model_start(
            serialized,
            messages,
            run_id=run_id,
            parent_run_id=parent_run_id,
            tags=tags,
            metadata=metadata,
            **kwargs,
        )

    def on_llm_new_token(
        self,
        token: str,
        *,
        chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        msg = self.llm_stream_message.get(str(run_id), None)
        if msg:
            self._run_sync(msg.stream_token(token))

        if self.stream_final_answer:
            self._append_to_last_tokens(token)

            if self.answer_reached:
                if not self.final_stream:
                    self.final_stream = Message(content="")
                self._run_sync(self.final_stream.stream_token(token))
                self.has_streamed_final_answer = True
            else:
                self.answer_reached = self._check_if_answer_reached()

        BaseTracer.on_llm_new_token(
            self,
            token,
            chunk=chunk,
            run_id=run_id,
            parent_run_id=parent_run_id,
            **kwargs,
        )

    def _start_trace(self, run: Run) -> None:
        super()._start_trace(run)
        context_var.set(self.context)

        if run.run_type in ["chain", "prompt"]:
            # Prompt templates are contained in chains or prompts (lcel)
            self._build_prompt(run.serialized or {}, run.inputs)

        ignore, parent_id = self._should_ignore_run(run)

        if ignore:
            return

        disable_human_feedback = not self._is_annotable(run)

        if run.run_type == "llm":
            msg = Message(
                id=run.id,
                content="",
                author=run.name,
                parent_id=parent_id,
                disable_human_feedback=disable_human_feedback,
            )
            self.llm_stream_message[str(run.id)] = msg
            self._run_sync(msg.send())
            return

        self._run_sync(
            Message(
                id=run.id,
                content="",
                author=run.name,
                parent_id=parent_id,
                disable_human_feedback=disable_human_feedback,
            ).send()
        )

    def _on_run_update(self, run: Run) -> None:
        """Process a run upon update."""
        context_var.set(self.context)

        ignore, parent_id = self._should_ignore_run(run)

        if ignore:
            return

        disable_human_feedback = not self._is_annotable(run)

        if run.run_type in ["chain"]:
            if self.prompt_sequence:
                self.prompt_sequence.pop()

        if run.run_type == "llm":
            provider, llm_settings = self._build_llm_settings(
                (run.serialized or {}), (run.extra or {}).get("invocation_params")
            )
            generations = (run.outputs or {}).get("generations", [])
            completion, language = self._get_completion(generations[0][0])
            current_prompt = (
                self.prompt_sequence.pop() if self.prompt_sequence else None
            )

            if current_prompt:
                current_prompt.provider = provider
                current_prompt.settings = llm_settings
                current_prompt.completion = completion
            else:
                generation_type = generations[0][0].get("type", "")
                current_prompt = self._build_default_prompt(
                    run, generation_type, provider, llm_settings, completion
                )

            msg = self.llm_stream_message.get(str(run.id), None)
            if msg:
                msg.content = completion
                msg.language = language
                msg.prompt = current_prompt
                self._run_sync(msg.update())

            if self.final_stream and self.has_streamed_final_answer:
                self.final_stream.content = completion
                self.final_stream.language = language
                self.final_stream.prompt = current_prompt
                self._run_sync(self.final_stream.send())
            return

        outputs = run.outputs or {}
        output_keys = list(outputs.keys())
        if output_keys:
            content = outputs.get(output_keys[0], "")
        else:
            return

        if run.run_type in ["agent", "chain"]:
            pass
            # # Add the response of the chain/tool
            # self._run_sync(
            #     Message(
            #         content=content,
            #         author=run.name,
            #         parent_id=parent_id,
            #         disable_human_feedback=disable_human_feedback,
            #     ).send()
            # )
        else:
            self._run_sync(
                Message(
                    id=run.id,
                    content=content,
                    author=run.name,
                    parent_id=parent_id,
                    disable_human_feedback=disable_human_feedback,
                ).update()
            )


LangchainCallbackHandler = LangchainTracer
AsyncLangchainCallbackHandler = LangchainTracer



import asyncio
import os

import click
import nest_asyncio
import uvicorn

nest_asyncio.apply()

from chainlit.auth import ensure_jwt_secret
from chainlit.cache import init_lc_cache
from chainlit.cli.utils import check_file
from chainlit.config import (
    BACKEND_ROOT,
    DEFAULT_HOST,
    DEFAULT_PORT,
    config,
    init_config,
    load_module,
)
from chainlit.logger import logger
from chainlit.markdown import init_markdown
from chainlit.secret import random_secret
from chainlit.server import app, max_message_size, register_wildcard_route_handler
from chainlit.telemetry import trace_event


# Create the main command group for Chainlit CLI
@click.group(context_settings={"auto_envvar_prefix": "CHAINLIT"})
@click.version_option(prog_name="Chainlit")
def cli():
    return


# Define the function to run Chainlit with provided options
def run_chainlit(target: str):
    host = os.environ.get("CHAINLIT_HOST", DEFAULT_HOST)
    port = int(os.environ.get("CHAINLIT_PORT", DEFAULT_PORT))
    config.run.host = host
    config.run.port = port

    check_file(target)
    # Load the module provided by the user
    config.run.module_name = target
    load_module(config.run.module_name)

    ensure_jwt_secret()

    register_wildcard_route_handler()

    # Create the chainlit.md file if it doesn't exist
    init_markdown(config.root)

    # Initialize the LangChain cache if installed and enabled
    init_lc_cache()

    log_level = "debug" if config.run.debug else "error"

    # Start the server
    async def start():
        config = uvicorn.Config(
            app,
            host=host,
            port=port,
            log_level=log_level,
            ws_max_size=max_message_size,
        )
        server = uvicorn.Server(config)
        await server.serve()

    # Run the asyncio event loop instead of uvloop to enable re entrance
    asyncio.run(start())
    # uvicorn.run(app, host=host, port=port, log_level=log_level)


# Define the "run" command for Chainlit CLI
@cli.command("run")
@click.argument("target", required=True, envvar="RUN_TARGET")
@click.option(
    "-w",
    "--watch",
    default=False,
    is_flag=True,
    envvar="WATCH",
    help="Reload the app when the module changes",
)
@click.option(
    "-h",
    "--headless",
    default=False,
    is_flag=True,
    envvar="HEADLESS",
    help="Will prevent to auto open the app in the browser",
)
@click.option(
    "-d",
    "--debug",
    default=False,
    is_flag=True,
    envvar="DEBUG",
    help="Set the log level to debug",
)
@click.option(
    "-c",
    "--ci",
    default=False,
    is_flag=True,
    envvar="CI",
    help="Flag to run in CI mode",
)
@click.option(
    "--no-cache",
    default=False,
    is_flag=True,
    envvar="NO_CACHE",
    help="Useful to disable third parties cache, such as langchain.",
)
@click.option("--host", help="Specify a different host to run the server on")
@click.option("--port", help="Specify a different port to run the server on")
def chainlit_run(target, watch, headless, debug, ci, no_cache, host, port):
    if host:
        os.environ["CHAINLIT_HOST"] = host
    if port:
        os.environ["CHAINLIT_PORT"] = port
    if ci:
        logger.info("Running in CI mode")

        config.project.enable_telemetry = False
        no_cache = True
        # This is required to have OpenAI LLM providers available for the CI run
        os.environ["OPENAI_API_KEY"] = "sk-FAKE-OPENAI-API-KEY"
        # This is required for authenticationt tests
        os.environ["CHAINLIT_AUTH_SECRET"] = "SUPER_SECRET"
    else:
        trace_event("chainlit run")

    config.run.headless = headless
    config.run.debug = debug
    config.run.no_cache = no_cache
    config.run.ci = ci
    config.run.watch = watch

    run_chainlit(target)


@cli.command("hello")
@click.argument("args", nargs=-1)
def chainlit_hello(args=None, **kwargs):
    trace_event("chainlit hello")
    hello_path = os.path.join(BACKEND_ROOT, "hello.py")
    run_chainlit(hello_path)


@cli.command("init")
@click.argument("args", nargs=-1)
def chainlit_init(args=None, **kwargs):
    trace_event("chainlit init")
    init_config(log=True)


@cli.command("create-secret")
@click.argument("args", nargs=-1)
def chainlit_create_secret(args=None, **kwargs):
    trace_event("chainlit secret")

    print(
        f"Copy the following secret into your .env file. Once it is set, changing it will logout all users with active sessions.\nCHAINLIT_AUTH_SECRET={random_secret()}"
    )



from typing import Any, Dict, List, Literal, Optional

from dataclasses_json import DataClassJsonMixin
from pydantic.dataclasses import dataclass


@dataclass
class BaseTemplate(DataClassJsonMixin):
    template: Optional[str] = None
    formatted: Optional[str] = None
    template_format: Optional[str] = "f-string"


@dataclass
class PromptMessage(BaseTemplate):
    # This is used for Langchain's MessagesPlaceholder
    placeholder_size: Optional[int] = None
    # This is used for OpenAI's function message
    name: Optional[str] = None
    role: Optional[Literal["system", "assistant", "user", "function"]] = None

    def to_openai(self):
        msg_dict = {"role": self.role, "content": self.formatted}
        if self.role == "function":
            msg_dict["name"] = self.name or ""
        return msg_dict

    def to_string(self):
        return f"{self.role}: {self.formatted}"


@dataclass
class Prompt(BaseTemplate):
    provider: Optional[str] = None
    id: Optional[str] = None
    inputs: Optional[Dict[str, str]] = None
    completion: Optional[str] = None
    settings: Optional[Dict[str, Any]] = None
    messages: Optional[List[PromptMessage]] = None
    functions: Optional[List[Dict]] = None
